@@ -266,7 +266,16 @@ static Bool LoadNvml(NvCtrlNvmlAttributes *nvml)
     GET_SYMBOL(_OPTIONAL, DeviceGetPowerUsage);
     GET_SYMBOL(_OPTIONAL, DeviceGetPowerManagementLimitConstraints);
     GET_SYMBOL(_OPTIONAL, DeviceGetPowerManagementDefaultLimit);
-
+    GET_SYMBOL(_OPTIONAL, DeviceGetThermalSettings);
+    GET_SYMBOL(_OPTIONAL, DeviceGetFanSpeedRPM);
+    GET_SYMBOL(_OPTIONAL, DeviceGetCoolerInfo);
+    GET_SYMBOL(_OPTIONAL, DeviceGetClockOffsets);
+    GET_SYMBOL(_OPTIONAL, DeviceSetClockOffsets);
+    GET_SYMBOL(_OPTIONAL, DeviceGetPerformanceModes);
+    GET_SYMBOL(_OPTIONAL, DeviceGetCurrentClockFreqs);
+    GET_SYMBOL(_OPTIONAL, DeviceGetPerformanceState);
+    GET_SYMBOL(_OPTIONAL, DeviceGetSupportedPerformanceStates);
+    GET_SYMBOL(_OPTIONAL, DeviceGetArchitecture);
 #undef GET_SYMBOL
 #undef EXPAND_STRING
 #undef STRINGIFY_SYMBOL
@@ -428,16 +437,11 @@ NvCtrlNvmlAttributes *NvCtrlInitNvmlAttributes(NvCtrlAttributePrivateHandle *h)
         nvmlDevice_t device;
         nvmlReturn_t ret = nvml->lib.DeviceGetHandleByIndex(devIdx, &device);
         if (ret == NVML_SUCCESS) {
-            unsigned int temp;
             unsigned int fans;
+            nvmlGpuThermalSettings_t pThermalSettings;
 
-            /*
-             * XXX Currently, NVML only allows to get the GPU temperature so
-             *     check for nvmlDeviceGetTemperature() success to figure
-             *     out if that sensor is available.
-             */
-            ret = nvml->lib.DeviceGetTemperature(device, NVML_TEMPERATURE_GPU,
-                                                 &temp);
+            ret = nvml->lib.DeviceGetThermalSettings(device, NVML_THERMAL_TARGET_ALL, //sensorIndex
+                                                     &pThermalSettings);
             if (ret == NVML_SUCCESS) {
                 if ((h->target_type == THERMAL_SENSOR_TARGET) &&
                     (h->target_id == nvml->sensorCount)) {
@@ -445,8 +449,8 @@ NvCtrlNvmlAttributes *NvCtrlInitNvmlAttributes(NvCtrlAttributePrivateHandle *h)
                     nvml->deviceIdx = devIdx;
                 }
 
-                nvml->sensorCountPerGPU[devIdx] = 1;
-                nvml->sensorCount++;
+                nvml->sensorCountPerGPU[devIdx] = pThermalSettings.count;
+                nvml->sensorCount += pThermalSettings.count;
             }
 
             ret = nvml->lib.DeviceGetNumFans(device, &fans);
@@ -599,7 +603,7 @@ static ReturnStatus NvCtrlNvmlGetGeneralStringAttribute(const CtrlTarget *ctrl_t
 static ReturnStatus NvCtrlNvmlGetGPUStringAttribute(const CtrlTarget *ctrl_target,
                                                     int attr, char **ptr)
 {
-    char res[MAX_NVML_STR_LEN];
+    char res[NVML_PERF_MODES_BUFFER_SIZE];
     const NvCtrlAttributePrivateHandle *h = getPrivateHandleConst(ctrl_target);
     const NvCtrlNvmlAttributes *nvml;
     nvmlDevice_t device;
@@ -650,9 +654,31 @@ static ReturnStatus NvCtrlNvmlGetGPUStringAttribute(const CtrlTarget *ctrl_targe
 
                 break;
             }
-            case NV_CTRL_STRING_SLI_MODE:
             case NV_CTRL_STRING_PERFORMANCE_MODES:
+            {
+                nvmlDevicePerfModes_t perfModes;
+
+                perfModes.version = nvmlDevicePerfModes_v1;
+                ret = nvml->lib.DeviceGetPerformanceModes(device, &perfModes);
+                if (ret == NVML_SUCCESS) {
+                    strcpy(res, perfModes.str);
+                }
+                break;
+            }
+
             case NV_CTRL_STRING_GPU_CURRENT_CLOCK_FREQS:
+            {
+                nvmlDeviceCurrentClockFreqs_t currentClockFreqs;
+
+                currentClockFreqs.version = nvmlDeviceCurrentClockFreqs_v1;
+                ret = nvml->lib.DeviceGetCurrentClockFreqs(device, &currentClockFreqs);
+                if (ret == NVML_SUCCESS) {
+                    strcpy(res, currentClockFreqs.str);
+                }
+                break;
+            }
+
+            case NV_CTRL_STRING_SLI_MODE:
             case NV_CTRL_STRING_MULTIGPU_MODE:
                 /*
                  * XXX We'll eventually need to add support for this attributes
@@ -812,7 +838,74 @@ ReturnStatus NvCtrlNvmlSetStringAttribute(CtrlTarget *ctrl_target,
     }
 }
 
+static nvmlReturn_t nvmlClockOffsetHelper(nvmlDevice_t device,
+                                          const NvCtrlNvmlAttributes *nvml,
+                                          int attr, int64_t *val,
+                                          CtrlAttributeValidValues *valid_values,
+                                          int *setVal)
+{
+    nvmlReturn_t ret;
 
+    switch (attr) {
+        case NV_CTRL_GPU_NVCLOCK_OFFSET:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+            {
+                nvmlDeviceArchitecture_t arch;
+
+                /* attributes are supported on Maxwell GPUs only */
+                ret = nvml->lib.DeviceGetArchitecture(device, &arch);
+                if ((ret != NVML_SUCCESS) || (arch != NVML_DEVICE_ARCH_MAXWELL)) {
+                    return NVML_ERROR_NOT_SUPPORTED;
+                }
+                /* fallthrough to case
+                 * NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS */
+                __attribute__((fallthrough));
+            }
+        case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+            {
+                nvmlClockOffset_t info;
+
+                switch (attr) {
+                    case NV_CTRL_GPU_NVCLOCK_OFFSET:
+                    case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+                        info.type = NVML_CLOCK_GRAPHICS;
+                        break;
+                    case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+                    case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+                        info.type = NVML_CLOCK_MEM;
+                        break;
+                }
+                info.version = nvmlClockOffset_v1;
+                info.pstate = NVML_PSTATE_0;
+                if (val || valid_values) {
+                    /* get current clock offset and valid values */
+                    ret = nvml->lib.DeviceGetClockOffsets(device, &info);
+
+                    if (ret != NVML_SUCCESS) {
+                        return ret;
+                    }
+                    if (val) {
+                        *val = info.clockOffsetMHz;
+                    }
+                    if (valid_values) {
+                        valid_values->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_RANGE;
+                        valid_values->range.min = info.minClockOffsetMHz;
+                        valid_values->range.max = info.maxClockOffsetMHz;
+                    }
+
+                    return NVML_SUCCESS;
+                }
+                if (setVal) {
+                    /* set new clock offset value */
+                    info.clockOffsetMHz = *setVal;
+                    ret = nvml->lib.DeviceSetClockOffsets(device, &info);
+                    return ret;
+                }
+            }
+    }
+    return NVML_SUCCESS;
+}
 
 /*
  * Get NVML Attribute Values
@@ -1072,11 +1165,45 @@ static ReturnStatus NvCtrlNvmlGetGPUAttribute(const CtrlTarget *ctrl_target,
                 }
                 break;
 
+            case NV_CTRL_GPU_CURRENT_PERFORMANCE_LEVEL:
+                {
+                    nvmlPstates_t pState;
+                    nvmlPstates_t pStates[NVML_MAX_GPU_PERF_PSTATES];
+                    nvmlReturn_t ret1;
+                    int i = 0;
+
+                    ret = nvml->lib.DeviceGetPerformanceState(device, &pState);
+                    ret1 = nvml->lib.DeviceGetSupportedPerformanceStates(device, pStates, NVML_MAX_GPU_PERF_PSTATES);
+                    if ((ret != NVML_SUCCESS) || (ret1 != NVML_SUCCESS)) {
+                        return NvCtrlNotSupported;
+                    }
+
+                    for (i = 0; i < NVML_MAX_GPU_PERF_PSTATES; i++) {
+                        if (pStates[i] == NVML_PSTATE_UNKNOWN) {
+                            continue;
+                        }
+                        if (pStates[i] == pState) {
+                            res = i;
+                            break;
+                        }
+                    }
+                }
+                break;
+
+            case NV_CTRL_GPU_NVCLOCK_OFFSET:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+            case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+                ret = nvmlClockOffsetHelper(device, nvml, attr, val, NULL, NULL);
+                if (ret == NVML_SUCCESS) {
+                    return NvCtrlSuccess;
+                }
+                break;
+
             case NV_CTRL_VIDEO_RAM:
             case NV_CTRL_GPU_PCIE_MAX_LINK_SPEED:
             case NV_CTRL_GPU_PCIE_CURRENT_LINK_SPEED:
             case NV_CTRL_BUS_TYPE:
-            case NV_CTRL_GPU_CURRENT_PERFORMANCE_LEVEL:
             case NV_CTRL_GPU_ADAPTIVE_CLOCK_STATE:
             case NV_CTRL_GPU_POWER_MIZER_MODE:
             case NV_CTRL_GPU_POWER_MIZER_DEFAULT_MODE:
@@ -1273,6 +1400,88 @@ static void getDeviceAndTargetIndex(const NvCtrlAttributePrivateHandle *h,
     return;
 }
 
+static void
+convertNvmlThermControllerToNvctrlSensorProvider(nvmlThermalController_t controller,
+                                                 unsigned int *sensorProvider)
+{
+    switch (controller) {
+        case NVML_THERMAL_CONTROLLER_NONE:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_NONE;
+            break;
+        case NVML_THERMAL_CONTROLLER_GPU_INTERNAL:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_GPU_INTERNAL;
+            break;
+        case NVML_THERMAL_CONTROLLER_ADM1032:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_ADM1032;
+            break;
+        case NVML_THERMAL_CONTROLLER_ADT7461:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_ADT7461;
+            break;
+        case NVML_THERMAL_CONTROLLER_MAX6649:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_MAX6649;
+            break;
+        case NVML_THERMAL_CONTROLLER_MAX1617:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_MAX1617;
+            break;
+        case NVML_THERMAL_CONTROLLER_LM99:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_LM99;
+            break;
+        case NVML_THERMAL_CONTROLLER_LM89:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_LM89;
+            break;
+        case NVML_THERMAL_CONTROLLER_LM64:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_LM64;
+            break;
+        case NVML_THERMAL_CONTROLLER_G781:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_G781;
+            break;
+        case NVML_THERMAL_CONTROLLER_ADT7473:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_ADT7473;
+            break;
+        case NVML_THERMAL_CONTROLLER_SBMAX6649:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_SBMAX6649;
+            break;
+        case NVML_THERMAL_CONTROLLER_VBIOSEVT:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_VBIOSEVT;
+            break;
+        case NVML_THERMAL_CONTROLLER_OS:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_OS;
+            break;
+        case NVML_THERMAL_CONTROLLER_UNKNOWN:
+        default:
+            *sensorProvider = NV_CTRL_THERMAL_SENSOR_PROVIDER_UNKNOWN;
+    }
+    return;
+}
+
+static void
+convertNvmlThermTargetToNvctrlThermTarget(nvmlThermalTarget_t target,
+                                          unsigned int *sensorTarget)
+{
+    switch (target) {
+        case NVML_THERMAL_TARGET_NONE:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_NONE;
+            break;
+        case NVML_THERMAL_TARGET_GPU:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_GPU;
+            break;
+        case NVML_THERMAL_TARGET_MEMORY:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_MEMORY;
+            break;
+        case NVML_THERMAL_TARGET_POWER_SUPPLY:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_POWER_SUPPLY;
+            break;
+        case NVML_THERMAL_TARGET_BOARD:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_BOARD;
+            break;
+        case NVML_THERMAL_TARGET_UNKNOWN:
+        default:
+            *sensorTarget = NV_CTRL_THERMAL_SENSOR_TARGET_UNKNOWN;
+            break;
+    }
+    return;
+}
+
 
 static ReturnStatus NvCtrlNvmlGetThermalAttribute(const CtrlTarget *ctrl_target,
                                                   int attr, int64_t *val)
@@ -1301,14 +1510,30 @@ static ReturnStatus NvCtrlNvmlGetThermalAttribute(const CtrlTarget *ctrl_target,
     if (ret == NVML_SUCCESS) {
         switch (attr) {
             case NV_CTRL_THERMAL_SENSOR_READING:
-               
             case NV_CTRL_THERMAL_SENSOR_PROVIDER:
             case NV_CTRL_THERMAL_SENSOR_TARGET:
-                /*
-                 * XXX We'll eventually need to add support for this attributes
-                 *     through NVML
-                 */
-                return NvCtrlNotSupported;
+                {
+                    nvmlGpuThermalSettings_t pThermalSettings;
+
+                    ret = nvml->lib.DeviceGetThermalSettings(device, NVML_THERMAL_TARGET_ALL, //sensorIndex
+                                                             &pThermalSettings);
+                    if (ret == NVML_SUCCESS) {
+                        switch (attr) {
+                            case NV_CTRL_THERMAL_SENSOR_READING:
+                                res = pThermalSettings.sensor[sensorId].currentTemp;
+                                break;
+                            case NV_CTRL_THERMAL_SENSOR_PROVIDER:
+                                convertNvmlThermControllerToNvctrlSensorProvider
+                                    (pThermalSettings.sensor[sensorId].controller, &res);
+                                break;
+                            case NV_CTRL_THERMAL_SENSOR_TARGET:
+                                convertNvmlThermTargetToNvctrlThermTarget
+                                    (pThermalSettings.sensor[sensorId].target, &res);
+                                break;
+                        }
+                    }
+                    break;
+                }
 
             default:
                 /* Did we forget to handle a sensor integer attribute? */
@@ -1329,6 +1554,50 @@ static ReturnStatus NvCtrlNvmlGetThermalAttribute(const CtrlTarget *ctrl_target,
     return NvCtrlNotSupported;
 }
 
+static void
+convertNvmlCoolerTargetToNvctrlCoolerTarget(nvmlCoolerTarget_t target,
+                                            unsigned int *coolerTarget)
+{
+    switch (target) {
+        case NVML_THERMAL_COOLER_TARGET_NONE:
+            *coolerTarget = NV_CTRL_THERMAL_COOLER_TARGET_NONE;
+            break;
+        case NVML_THERMAL_COOLER_TARGET_GPU:
+            *coolerTarget = NV_CTRL_THERMAL_COOLER_TARGET_GPU;
+            break;
+        case NVML_THERMAL_COOLER_TARGET_MEMORY:
+            *coolerTarget = NV_CTRL_THERMAL_COOLER_TARGET_MEMORY;
+            break;
+        case NVML_THERMAL_COOLER_TARGET_POWER_SUPPLY:
+            *coolerTarget = NV_CTRL_THERMAL_COOLER_TARGET_POWER_SUPPLY;
+            break;
+        case NVML_THERMAL_COOLER_TARGET_GPU_RELATED:
+        default:
+            *coolerTarget = NV_CTRL_THERMAL_COOLER_TARGET_GPU_RELATED;
+            break;
+    }
+    return;
+}
+
+static void
+convertNvmlCoolerSignalToNvctrlCoolerControlType(nvmlCoolerControl_t signalType,
+                                                 unsigned int *coolerControlType)
+{
+    switch (signalType) {
+        case NVML_THERMAL_COOLER_SIGNAL_NONE:
+            *coolerControlType = NV_CTRL_THERMAL_COOLER_CONTROL_TYPE_NONE;
+            break;
+        case NVML_THERMAL_COOLER_SIGNAL_TOGGLE:
+            *coolerControlType = NV_CTRL_THERMAL_COOLER_CONTROL_TYPE_TOGGLE;
+            break;
+        case NVML_THERMAL_COOLER_SIGNAL_VARIABLE:
+        default:
+            *coolerControlType = NV_CTRL_THERMAL_COOLER_CONTROL_TYPE_VARIABLE;
+            break;
+    }
+    return;
+}
+
 static ReturnStatus NvCtrlNvmlGetCoolerAttribute(const CtrlTarget *ctrl_target,
                                                  int attr, int64_t *val)
 {
@@ -1360,15 +1629,37 @@ static ReturnStatus NvCtrlNvmlGetCoolerAttribute(const CtrlTarget *ctrl_target,
             case NV_CTRL_THERMAL_COOLER_CURRENT_LEVEL:
                 ret = nvml->lib.DeviceGetFanSpeed_v2(device, coolerId, &res);
                 break;
-
-            case NV_CTRL_THERMAL_COOLER_SPEED:
-            case NV_CTRL_THERMAL_COOLER_CONTROL_TYPE:
             case NV_CTRL_THERMAL_COOLER_TARGET:
-                /*
-                 * XXX We'll eventually need to add support for this attributes
-                 *     through NVML
-                 */
-                return NvCtrlNotSupported;
+            case NV_CTRL_THERMAL_COOLER_CONTROL_TYPE:
+                {
+                    nvmlCoolerInfo_t coolerInfo;
+                    coolerInfo.version = nvmlCoolerInfo_v1;
+                    coolerInfo.index = coolerId;
+                    ret = nvml->lib.DeviceGetCoolerInfo(device, &coolerInfo);
+                    if (ret == NVML_SUCCESS) {
+                        switch (attr) {
+                            case NV_CTRL_THERMAL_COOLER_CONTROL_TYPE:
+                                convertNvmlCoolerSignalToNvctrlCoolerControlType(coolerInfo.signalType, &res);
+                                break;
+                            case NV_CTRL_THERMAL_COOLER_TARGET:
+                                convertNvmlCoolerTargetToNvctrlCoolerTarget(coolerInfo.target, &res);
+                                break;
+                        }
+                    }
+                }
+                break;
+            case NV_CTRL_THERMAL_COOLER_SPEED:
+                {
+                    nvmlFanSpeedInfo_t fanSpeed;
+                    fanSpeed.version = nvmlFanSpeedInfo_v1;
+                    fanSpeed.fan = coolerId;
+                    ret = nvml->lib.DeviceGetFanSpeedRPM(device, &fanSpeed);
+                    if (ret == NVML_SUCCESS) {
+                        *val = fanSpeed.speed;
+                        return NvCtrlSuccess;
+                    }
+                }
+                break;
 
             default:
                 /* Did we forget to handle a cooler integer attribute? */
@@ -1517,6 +1808,13 @@ static ReturnStatus NvCtrlNvmlSetGPUAttribute(CtrlTarget *ctrl_target,
                 }
                 break;
 
+            case NV_CTRL_GPU_NVCLOCK_OFFSET:
+            case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+                ret = nvmlClockOffsetHelper(device, nvml, attr, NULL, NULL, &val);
+                break;
+
             case NV_CTRL_GPU_CURRENT_CLOCK_FREQS:
             case NV_CTRL_GPU_POWER_MIZER_MODE:
             case NV_CTRL_DITHERING:
@@ -1547,6 +1845,9 @@ static ReturnStatus NvCtrlNvmlSetGPUAttribute(CtrlTarget *ctrl_target,
 
     /* An NVML error occurred */
     printNvmlError(ret);
+    if (ret == NVML_ERROR_NO_PERMISSION) {
+        return NvCtrlNoPermission;
+    }
     return NvCtrlNotSupported;
 }
 
@@ -1597,6 +1898,9 @@ static ReturnStatus NvCtrlNvmlSetCoolerAttribute(CtrlTarget *ctrl_target,
 
     /* An NVML error occurred */
     printNvmlError(ret);
+    if (ret == NVML_ERROR_NO_PERMISSION) {
+        return NvCtrlNoPermission;
+    }
     return NvCtrlNotSupported;
 }
 
@@ -1726,6 +2030,46 @@ NvCtrlNvmlGetGPUBinaryAttribute(const CtrlTarget *ctrl_target,
                 }
                 break;
             }
+            case NV_CTRL_BINARY_DATA_THERMAL_SENSORS_USED_BY_GPU:
+            {
+                unsigned int *sensor_data;
+                unsigned int count = 1;
+                int offset = 0;
+                int i = 0;
+                nvmlGpuThermalSettings_t pThermalSettings;
+
+                ret = nvml->lib.DeviceGetThermalSettings(device,
+                                                         NVML_THERMAL_TARGET_ALL,
+                                                         &pThermalSettings);
+                if (ret != NVML_SUCCESS) {
+                    return NvCtrlNotSupported;
+                }
+
+                count = pThermalSettings.count;
+
+                /*
+                 * The format of the returned data is:
+                 *
+                 * int       unsigned int number of SENSORS
+                 * int * n   unsigned int SENSOR indices
+                 */
+
+                *len = (count + 1) * sizeof(unsigned int);
+                sensor_data = (unsigned int *) nvalloc(*len);
+                memset(sensor_data, 0, *len);
+                *data = (unsigned char *) sensor_data;
+
+                /* Calculate global sensor index offset for this GPU */
+                for (i = 0; i < nvml->deviceIdx; i++) {
+                    offset += nvml->sensorCountPerGPU[i];
+                }
+
+                sensor_data[0] = count;
+                for (i = 0; i < count; i++) {
+                    sensor_data[i+1] = i + offset;
+                }
+                break;
+            }
             case NV_CTRL_BINARY_DATA_GPU_ECC_DETAILED_ERRORS_SINGLE_BIT:
                 ret = getDeviceMemoryCounts(ctrl_target, nvml, device,
                                             NVML_MEMORY_ERROR_TYPE_CORRECTED,
@@ -1751,7 +2095,6 @@ NvCtrlNvmlGetGPUBinaryAttribute(const CtrlTarget *ctrl_target,
                                             data, len);
                 break;
             case NV_CTRL_BINARY_DATA_FRAMELOCKS_USED_BY_GPU:
-            case NV_CTRL_BINARY_DATA_THERMAL_SENSORS_USED_BY_GPU:
             case NV_CTRL_BINARY_DATA_DISPLAYS_CONNECTED_TO_GPU:
             case NV_CTRL_BINARY_DATA_DISPLAYS_ON_GPU:
             case NV_CTRL_BINARY_DATA_GPU_FLAGS:
@@ -1840,13 +2183,13 @@ NvCtrlNvmlGetGPUValidStringAttributeValues(int attr,
         case NV_CTRL_STRING_PRODUCT_NAME:
         case NV_CTRL_STRING_VBIOS_VERSION:
         case NV_CTRL_STRING_GPU_UUID:
+        case NV_CTRL_STRING_GPU_CURRENT_CLOCK_FREQS:
+        case NV_CTRL_STRING_PERFORMANCE_MODES:
             val->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_STRING;
             return NvCtrlSuccess;
 
         case NV_CTRL_STRING_SLI_MODE:
-        case NV_CTRL_STRING_PERFORMANCE_MODES:
         case NV_CTRL_STRING_MULTIGPU_MODE:
-        case NV_CTRL_STRING_GPU_CURRENT_CLOCK_FREQS:
         case NV_CTRL_STRING_GPU_UTILIZATION:
             /*
              * XXX We'll eventually need to add support for this attributes. For
@@ -1990,9 +2333,17 @@ NvCtrlNvmlGetGPUValidAttributeValues(const CtrlTarget *ctrl_target, int attr,
             case NV_CTRL_ATTR_NVML_GPU_GET_POWER_USAGE:
             case NV_CTRL_ATTR_NVML_GPU_MAX_TGP:
             case NV_CTRL_ATTR_NVML_GPU_DEFAULT_TGP:
+            case NV_CTRL_GPU_CURRENT_PERFORMANCE_LEVEL:
                 val->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_INTEGER;
                 break;
 
+            case NV_CTRL_GPU_NVCLOCK_OFFSET:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+            case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+            case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+                ret = nvmlClockOffsetHelper(device, nvml, attr, NULL, val, NULL);
+                break;
+
             case NV_CTRL_GPU_ECC_SUPPORTED:
             case NV_CTRL_GPU_ECC_STATUS:
             case NV_CTRL_GPU_ECC_CONFIGURATION:
@@ -2020,7 +2371,6 @@ NvCtrlNvmlGetGPUValidAttributeValues(const CtrlTarget *ctrl_target, int attr,
             case NV_CTRL_GPU_PCIE_MAX_LINK_SPEED:
             case NV_CTRL_GPU_PCIE_CURRENT_LINK_SPEED:
             case NV_CTRL_BUS_TYPE:
-            case NV_CTRL_GPU_CURRENT_PERFORMANCE_LEVEL:
             case NV_CTRL_GPU_ADAPTIVE_CLOCK_STATE:
             case NV_CTRL_GPU_POWER_MIZER_MODE:
             case NV_CTRL_GPU_POWER_MIZER_DEFAULT_MODE:
@@ -2106,14 +2456,22 @@ NvCtrlNvmlGetThermalValidAttributeValues(const CtrlTarget *ctrl_target,
     if (ret == NVML_SUCCESS) {
         switch (attr) {
             case NV_CTRL_THERMAL_SENSOR_READING:
+                {
+                    nvmlGpuThermalSettings_t pThermalSettings;
+
+                    ret = nvml->lib.DeviceGetThermalSettings(device, NVML_THERMAL_TARGET_ALL, //sensorIndex
+                                                             &pThermalSettings);
+                    if (ret == NVML_SUCCESS) {
+                        val->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_RANGE;
+                        val->range.min = pThermalSettings.sensor[sensorId].defaultMinTemp;
+                        val->range.max = pThermalSettings.sensor[sensorId].defaultMaxTemp;
+                    }
+                    break;
+                }
             case NV_CTRL_THERMAL_SENSOR_PROVIDER:
             case NV_CTRL_THERMAL_SENSOR_TARGET:
-                /*
-                 * XXX We'll eventually need to add support for this attributes
-                 *     through NVML
-                 */
-                return NvCtrlAttributeNotAvailable;
-
+                val->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_INTEGER;
+                break;
             default:
                 /* The attribute queried is not sensor-targeted */
                 return NvCtrlAttributeNotAvailable;
@@ -2177,11 +2535,8 @@ NvCtrlNvmlGetCoolerValidAttributeValues(const CtrlTarget *ctrl_target,
             case NV_CTRL_THERMAL_COOLER_SPEED:
             case NV_CTRL_THERMAL_COOLER_CONTROL_TYPE:
             case NV_CTRL_THERMAL_COOLER_TARGET:
-                /*
-                 * XXX We'll eventually need to add support for this attributes
-                 *     through NVML
-                 */
-                return NvCtrlAttributeNotAvailable;
+                val->valid_type = CTRL_ATTRIBUTE_VALID_TYPE_INTEGER;
+                break;
 
             default:
                 /* The attribute queried is not fan-targeted */
@@ -2261,6 +2616,10 @@ static ReturnStatus NvCtrlNvmlGetIntegerAttributePerms(int attr,
 {
     /* Set write permissions */
     switch (attr) {
+        case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+        case NV_CTRL_GPU_NVCLOCK_OFFSET:
         case NV_CTRL_GPU_COOLER_MANUAL_CONTROL:
         case NV_CTRL_THERMAL_COOLER_LEVEL:
         case NV_CTRL_GPU_ECC_CONFIGURATION:
@@ -2304,19 +2663,36 @@ static ReturnStatus NvCtrlNvmlGetIntegerAttributePerms(int attr,
         case NV_CTRL_GPU_ECC_AGGREGATE_SINGLE_BIT_ERRORS:
         case NV_CTRL_GPU_ECC_DOUBLE_BIT_ERRORS:
         case NV_CTRL_GPU_ECC_AGGREGATE_DOUBLE_BIT_ERRORS:
+        case NV_CTRL_GPU_NVCLOCK_OFFSET_ALL_PERFORMANCE_LEVELS:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET_ALL_PERFORMANCE_LEVELS:
+        case NV_CTRL_GPU_MEM_TRANSFER_RATE_OFFSET:
+        case NV_CTRL_GPU_NVCLOCK_OFFSET:
         /* CTRL_ATTRIBUTE_VALID_TYPE_INT_BITS */
+            perms->read = NV_TRUE;
+            perms->valid_targets = CTRL_TARGET_PERM_BIT(GPU_TARGET);
+            break;
         /* GPU_TARGET non-readable attribute */
         case NV_CTRL_GPU_ECC_RESET_ERROR_STATUS:
             perms->read  = NV_FALSE;
             perms->valid_targets = CTRL_TARGET_PERM_BIT(GPU_TARGET);
             break;
 
+        case NV_CTRL_THERMAL_COOLER_CONTROL_TYPE:
+        case NV_CTRL_THERMAL_COOLER_TARGET:
+        case NV_CTRL_THERMAL_COOLER_SPEED:
         case NV_CTRL_THERMAL_COOLER_LEVEL:
         case NV_CTRL_THERMAL_COOLER_CURRENT_LEVEL:
             perms->valid_targets = CTRL_TARGET_PERM_BIT(COOLER_TARGET);
             perms->read = NV_TRUE;
             break;
 
+        case NV_CTRL_THERMAL_SENSOR_READING:
+        case NV_CTRL_THERMAL_SENSOR_PROVIDER:
+        case NV_CTRL_THERMAL_SENSOR_TARGET:
+            perms->valid_targets = CTRL_TARGET_PERM_BIT(THERMAL_SENSOR_TARGET);
+            perms->read = NV_TRUE;
+            break;
+
         default:
             perms->valid_targets = 0;
             perms->read = NV_FALSE;
@@ -2336,6 +2712,8 @@ static ReturnStatus NvCtrlNvmlGetStringAttributePerms(int attr,
         case NV_CTRL_STRING_PRODUCT_NAME:
         case NV_CTRL_STRING_VBIOS_VERSION:
         case NV_CTRL_STRING_GPU_UUID:
+        case NV_CTRL_STRING_GPU_CURRENT_CLOCK_FREQS:
+        case NV_CTRL_STRING_PERFORMANCE_MODES:
             perms->valid_targets = CTRL_TARGET_PERM_BIT(GPU_TARGET);
             perms->read = NV_TRUE;
             break;
